@@ -117,6 +117,8 @@ glade_xml_init (GladeXML *self)
     priv->tree = NULL;
     priv->tooltips = gtk_tooltips_new();
     gtk_tooltips_enable(priv->tooltips);
+    gtk_object_ref(GTK_OBJECT(priv->tooltips));
+    gtk_object_sink(GTK_OBJECT(priv->tooltips));
     priv->name_hash = g_hash_table_new(g_str_hash, g_str_equal);
     priv->signals = g_hash_table_new(g_str_hash, g_str_equal);
     priv->radio_groups = g_hash_table_new (g_str_hash, g_str_equal);
@@ -647,11 +649,14 @@ glade_xml_set_toplevel(GladeXML *xml, GtkWindow *window)
 	glade_xml_pop_accel(xml);
     /* maybe put an assert that xml->priv->accel_groups == NULL here? */
     xml->priv->accel_groups = NULL;
-    /* the window should hold a reference to the tooltips object */
-    gtk_object_ref(GTK_OBJECT(xml->priv->tooltips));
-    g_object_set_qdata_full(G_OBJECT(window), glade_xml_tooltips_id,
-			    xml->priv->tooltips,
-			    (GDestroyNotify)gtk_object_unref);
+
+    if (GTK_IS_WINDOW (window)) {
+	/* the window should hold a reference to the tooltips object */
+	gtk_object_ref(GTK_OBJECT(xml->priv->tooltips));
+	g_object_set_qdata_full(G_OBJECT(window), glade_xml_tooltips_id,
+				xml->priv->tooltips,
+				(GDestroyNotify)gtk_object_unref);
+    }
 }
 
 /**
@@ -811,7 +816,10 @@ glade_xml_add_atk_actions(GladeXML *xml, GtkWidget *w, GladeWidgetInfo *info)
 					    j,
 					    action_info->description);	
 	    } else {
-		    g_warning("could not find action named %s", action_info->action_name);
+		/* don't show a warning -- the action names might not
+                 * be registered if libgail hasn't been loaded */
+
+		/* g_warning("could not find action named %s", action_info->action_name); */
 	    }
         }
     }	
@@ -853,7 +861,7 @@ glade_xml_add_atk_relations(GladeXML *xml, GtkWidget *w, GladeWidgetInfo *info)
 	GtkWidget *target_widget = glade_xml_get_widget (xml, rinfo->target);
 	AtkRelationType relation_type;
 
-	relation_type = atk_relation_type_from_string (rinfo->type);
+	relation_type = atk_relation_type_for_name (rinfo->type);
 	if (target_widget) {
 	    AtkObject *target_accessible;
 
@@ -1062,6 +1070,37 @@ glade_create_custom(GladeXML *xml, gchar *func_name, gchar *name,
 				  int1, int2, custom_user_data);
 }
 
+static GtkWidget *
+custom_new (GladeXML *xml, GladeWidgetInfo *info)
+{
+    GtkWidget *wid = NULL;
+    gchar *func_name = NULL, *string1 = NULL, *string2 = NULL;
+    gint int1 = 0, int2 = 0;
+    int i;
+    
+    for (i = 0; i < info->n_properties; i++) {
+	char *name  = info->properties[i].name;
+	char *value = info->properties[i].value;
+	
+	if (!strcmp(name, "creation_function"))
+	    func_name = value;
+	else if (!strcmp(name, "string1"))
+	    string1 = value;
+	else if (!strcmp(name, "string2"))
+	    string2 = value;
+	else if (!strcmp(name, "int1"))
+	    int1 = strtol(value, NULL, 0);
+	else if (!strcmp(name, "int2"))
+	    int2 = strtol(value, NULL, 0);
+    }
+    wid = glade_create_custom(xml, func_name, info->name, string1,
+			      string2, int1, int2);
+    /* fallback to prevent segfault */
+    if (wid == NULL)
+	wid = gtk_label_new("[custom widget creation failed]");
+    return wid;
+}
+
 /**
  * glade_enum_from_string
  * @type: the GType for this enum type.
@@ -1152,8 +1191,7 @@ glade_flags_from_string (GType type, const char *string)
 
 	    if (eos)
 		break;
-	} else
-	    i++;
+	}
     }
     
     g_free (flagstr);
@@ -1179,21 +1217,11 @@ glade_xml_build_interface(GladeXML *self, GladeInterface *iface,
 	wid = g_hash_table_lookup(iface->names, root);
 	g_return_if_fail(wid != NULL);
 	w = glade_xml_build_widget(self, wid);
-	if (GTK_IS_WINDOW(w)) {
-	    if (self->priv->focus_widget)
-		gtk_widget_grab_focus(self->priv->focus_widget);
-	    if (self->priv->default_widget)
-		gtk_widget_grab_default(self->priv->default_widget);
-	}
     } else {
 	/* build all toplevel nodes */
 	for (i = 0; i < iface->n_toplevels; i++) {
 	    w = glade_xml_build_widget(self, iface->toplevels[i]);
 	}
-	if (self->priv->focus_widget)
-	    gtk_widget_grab_focus(self->priv->focus_widget);
-	if (self->priv->default_widget)
-	    gtk_widget_grab_default(self->priv->default_widget);
     }
 }
 
@@ -1276,6 +1304,128 @@ get_build_data(GType type)
 	return &widget_build_data;
 }
 
+
+/* used to store a GArray of custom property handlers */
+static GQuark       glade_custom_props_id = 0;
+static const gchar *glade_custom_props_key = "libglade::custom-props";
+/* used to store a cache of the property handlers (ie. this type's
+ * properties + parent properties) */
+static GQuark       glade_custom_props_cache_id = 0;
+static const gchar *glade_custom_props_cache_key = "libglade::custom-props-cache";
+typedef struct _CustomPropInfo CustomPropInfo;
+struct _CustomPropInfo {
+    GQuark name_quark;
+    GladeApplyCustomPropFunc apply_prop;
+};
+
+/* utility function for invalidating the cached property key */
+static void
+invalidate_custom_prop_cache(GType type)
+{
+    gpointer data;
+    GType *children;
+    guint i, n_children = 0;
+
+    data = g_type_get_qdata(type, glade_custom_props_cache_id);
+    /* if there is no cached custom prop data, then no children will
+     * have cached data. */
+    if (!data)
+	return;
+
+    g_type_set_qdata(type, glade_custom_props_cache_id, NULL);
+    g_free(data);
+    children = g_type_children(type, &n_children);
+    for (i = 0; i < n_children; i++)
+	invalidate_custom_prop_cache(children[i]);
+    g_free(children);
+}
+
+void
+glade_register_custom_prop(GType type, const gchar *prop_name,
+			   GladeApplyCustomPropFunc apply_prop)
+{
+    GArray *array;
+    CustomPropInfo prop_info;
+    
+    /* set up quarks ... */
+    if (glade_custom_props_id == 0) {
+	glade_custom_props_id = g_quark_from_static_string(glade_custom_props_key);
+	glade_custom_props_cache_id = g_quark_from_static_string(glade_custom_props_cache_key);
+    }
+
+    /* get the GArray of prop info */
+    array = g_type_get_qdata(type, glade_custom_props_id);
+    if (!array) {
+	array = g_array_new(TRUE, FALSE, sizeof(CustomPropInfo));
+	g_type_set_qdata(type, glade_custom_props_id, array);
+    }
+
+    /* append the extra prop */
+    prop_info.name_quark = g_quark_from_string(prop_name);
+    prop_info.apply_prop = apply_prop;
+    g_array_append_val(array, prop_info);
+
+    /* invalidate cached info, if any */
+    invalidate_custom_prop_cache(type);
+}
+
+static CustomPropInfo *
+get_custom_prop_info(GType type)
+{
+    CustomPropInfo *prop_info, *parent_info = NULL;
+    GType parent;
+    GArray *array;
+    gint length;
+
+    /* if glade_register_custom_prop hasn't been called ... */
+    if (glade_custom_props_id == 0) {
+	return NULL;
+    }
+
+    /* has the prop info been calculated/cached already? */
+    prop_info = g_type_get_qdata(type, glade_custom_props_cache_id);
+    if (prop_info)
+	return prop_info;
+
+    array = g_type_get_qdata(type, glade_custom_props_id);
+    parent = g_type_parent(type);
+    prop_info = NULL;
+    length = 0;
+    if (parent == G_TYPE_INVALID) {/* hit the top */
+	if (array) {
+	    length = array->len;
+	    prop_info = g_memdup(array->data,
+				 sizeof(CustomPropInfo) * (length + 1));
+	}
+    } else { /* merge parent info with our info */
+	gint count = 0;
+
+	parent_info = get_custom_prop_info(parent);
+	if (parent_info)
+	    while (parent_info[count].name_quark != 0)
+		count++;
+	length = count;
+	if (array) length += array->len;
+	prop_info = g_new(CustomPropInfo, length + 1);
+	/* concatenate info */
+	if (count > 0)
+	    memcpy(prop_info, parent_info,
+		   sizeof(CustomPropInfo) * count);
+	if (array)
+	    memcpy(&prop_info[count], array->data,
+		   sizeof(CustomPropInfo) * array->len);
+    }
+    /* make sure it is null terminated */
+    if (prop_info != NULL) {
+	prop_info[length].name_quark = 0;
+	prop_info[length].apply_prop = NULL;
+    }
+
+    g_type_set_qdata(type, glade_custom_props_cache_id, prop_info);
+    return prop_info;
+}
+
+
 /**
  * glade_xml_set_value_from_string
  * @xml: the GladeXML object.
@@ -1313,7 +1463,10 @@ glade_xml_set_value_from_string (GladeXML *xml,
 	g_value_set_uchar(value, (guchar)string[0]);
 	break;
     case G_TYPE_BOOLEAN:
-	g_value_set_boolean(value, string[0] == 'T' || string[0] == 'y');
+	g_value_set_boolean(value, 
+			    g_ascii_tolower (string[0]) == 't'
+			    || g_ascii_tolower (string[0]) == 'y'
+			    || strtol (string, NULL, 0));
 	break;
     case G_TYPE_INT:
 	g_value_set_int(value, strtol(string, NULL, 0));
@@ -1404,6 +1557,13 @@ glade_xml_set_value_from_string (GladeXML *xml,
     return ret;
 }
 
+/* a small struct to hold info about a custom prop until we apply it */
+typedef struct _CustomPropData CustomPropData;
+struct _CustomPropData {
+    GladeApplyCustomPropFunc apply_prop;
+    GladeProperty *prop;
+};
+
 /**
  * glade_standard_build_widget
  * @xml: the GladeXML object.
@@ -1422,22 +1582,47 @@ glade_standard_build_widget(GladeXML *xml, GType widget_type,
 			    GladeWidgetInfo *info)
 {
     static GArray *props_array = NULL;
+    static GArray *custom_props_array = NULL;
     GObjectClass *oclass;
+    CustomPropInfo *custom_props;
     GtkWidget *widget;
     GList *deferred_props = NULL, *tmp;
     guint i;
 
-    if (!props_array)
+    if (!props_array) {
 	props_array = g_array_new(FALSE, FALSE, sizeof(GParameter));
+	custom_props_array = g_array_new(FALSE, FALSE, sizeof(CustomPropData));
+    }
 
     /* we ref the class here as a slight optimisation */
     oclass = g_type_class_ref(widget_type);
 
+    custom_props = get_custom_prop_info(widget_type);
+
     /* collect properties */
     for (i = 0; i < info->n_properties; i++) {
+	GQuark name_quark;
 	GParameter param = { NULL };
 	GParamSpec *pspec;
 
+	if (custom_props &&
+	    (name_quark = g_quark_try_string(info->properties[i].name)) != 0) {
+	    gint j = 0;
+
+	    while (custom_props[j].name_quark) {
+		if (custom_props[j].name_quark == name_quark) {
+		    CustomPropData prop_data;
+
+		    prop_data.apply_prop = custom_props[j].apply_prop;
+		    prop_data.prop = &info->properties[i];
+		    g_array_append_val(custom_props_array, prop_data);
+		    break;
+		}
+		j++;
+	    }
+	    if (custom_props[j].name_quark != 0) /* a prop was matched */
+		continue;
+	}
 	pspec = g_object_class_find_property(oclass, info->properties[i].name);
 	if (!pspec) {
 	    g_warning("unknown property `%s' for class `%s'",
@@ -1475,6 +1660,16 @@ glade_standard_build_widget(GladeXML *xml, GType widget_type,
 	g_value_unset(&g_array_index(props_array, GParameter, i).value);
     }
 
+    /* do custom props */
+    for (i = 0; i < custom_props_array->len; i++) {
+	CustomPropData *data;
+
+	data = &g_array_index(custom_props_array, CustomPropData, i);
+	if (data->apply_prop)
+	    (* data->apply_prop) (xml, widget, data->prop->name,
+				  data->prop->value);
+    }
+
     /* handle deferred properties */
     for (tmp = deferred_props; tmp; tmp = tmp->next) {
 	GladeProperty *prop = tmp->data;
@@ -1484,6 +1679,7 @@ glade_standard_build_widget(GladeXML *xml, GType widget_type,
     g_list_free(tmp);
 
     g_array_set_size(props_array, 0);
+    g_array_set_size(custom_props_array, 0);
     g_type_class_unref(oclass);
 
     return widget;
@@ -1500,7 +1696,7 @@ glade_standard_build_widget(GladeXML *xml, GType widget_type,
  * This sets the packing property on container @parent of widget
  * @child with @name to @value
  **/
-static void
+void
 glade_xml_set_packing_property (GladeXML   *self,
 				GtkWidget  *parent, GtkWidget  *child,
 				const char *name,   const char *value)
@@ -1564,7 +1760,7 @@ glade_standard_build_children(GladeXML *self, GtkWidget *parent,
 		self, parent, child,
 		info->children[i].properties[j].name,
 		info->children[i].properties[j].value);
-
+	
 	gtk_widget_thaw_child_notify(child);
 	g_object_unref(G_OBJECT(child));
     }
@@ -1644,27 +1840,35 @@ glade_xml_build_widget(GladeXML *self, GladeWidgetInfo *info)
 {
     GType type;
     GtkWidget *ret;
-
-    debug(g_message("Widget class: %s", info->class));
+    
+    debug(g_message("Widget class: %s\tname: %s", info->class, info->name));
     if (!strcmp(info->class, "Placeholder")) {
 	static int warned = 0;
 	if (!warned++)
 	    g_warning("placeholders exist in interface description");
 	ret = gtk_label_new("[placeholder]");
-	gtk_widget_show(ret);
-	return ret;
-    }
-    type = g_type_from_name(info->class);
-    if (type == G_TYPE_INVALID) {
-	char buf[50];
-	g_warning("unknown widget class '%s'", info->class);
-	g_snprintf(buf, 49, "[a %s]", info->class);
-	ret = gtk_label_new(buf);
-	gtk_widget_show(ret);
+    } else if (!strcmp (info->class, "Custom")) {
+	ret = custom_new (self, info);
     } else {
-	ret = get_build_data(type)->new(self, type, info);
+	type = g_type_from_name(info->class);
+	if (type == G_TYPE_INVALID) {
+	    char buf[50];
+	    g_warning("unknown widget class '%s'", info->class);
+	    g_snprintf(buf, 49, "[a %s]", info->class);
+	    ret = gtk_label_new(buf);
+	} else {
+	    ret = get_build_data(type)->new(self, type, info);
+	}
     }
+
+    if (GTK_IS_WINDOW (ret))
+	glade_xml_set_toplevel (self, GTK_WINDOW (ret));
+
     glade_xml_set_common_params(self, ret, info);
+
+    if (GTK_IS_WINDOW (ret))
+	glade_xml_set_toplevel (self, NULL);
+
     return ret;
 }
 
@@ -1818,8 +2022,13 @@ glade_xml_set_common_params(GladeXML *self, GtkWidget *widget,
 	}
     }
 
-    if (data && data->build_children && info->children)
-	data->build_children(self, widget, info);
+    if (data && data->build_children && info->children) {
+	if (GTK_IS_CONTAINER (widget))
+	    data->build_children(self, widget, info);
+	else
+	    g_warning ("widget %s (%s) has children, but is not a GtkContainer.",
+		       info->name, g_type_name (G_TYPE_FROM_INSTANCE (widget)));
+    }
 }
 
 gchar **
